#if !defined(PARSER_HH)
#define PARSER_HH

#include <iostream>

#include "Lexer.hh"
#include "AstManager.hh"
#include "ErrorManager.hh"

///////////////////////////////////////////////////////////////////////////
class Parser {
///////////////////////////////////////////////////////////////////////////

public :

  Parser( Lexer& lex, AstManager& ast, ErrorManager& error ) : 
    lex_(lex), ast_(ast), error_(error) { }
  ~Parser( void ) { }
  
  AstNode* operator()( void ) {
    AstNode* expression = 0;
    // Advances the token cursor to the first token
    lex_.next();
    expression = parseExpression_();
    // Expects to find the token cursor on top of EOF
    if (lex_.cur().getPartOfSpeech() != Token::EOFF) {
      error_("Garbage after end of program");
    }
    return expression;
  }  
  
private :
  Lexer&        lex_;
  AstManager&   ast_;
  ErrorManager& error_;
  
  AstNode* parseExpression_() {
    AstNode* etemp = 0;
    // If the current token is a parenthesis ...
    if (lex_.cur().getPartOfSpeech() == Token::LPAR) {
      // ... then this is a parenthesized expression, ...
      etemp = parseParenExpression_();
    } else if (lex_.cur().getPartOfSpeech() == Token::NUMBER) {
      // ...otherwise it is a single number
      etemp = ast_.newOperand();
      etemp->setValue(atoi(lex_.cur().getLexeme().data()));
    }
    // Check if there is more legal stuff afterwards
    return parseRestExpression_(etemp);
  }

  AstNode* parseRestExpression_(AstNode* term) {
    AstNode* e = 0;
    // Advance the token cursor...
    lex_.next();
    if (lex_.cur().getPartOfSpeech() == Token::OPERATOR) {
      // ... if an operator is found, then there is a trailing expression, ...
      e = ast_.newOperator();
      // term was parsed beforehand
      e->setOperand(AstNode::LEFT, term);
      e->setOpType(parseOpType_());
      // Advances the token cursor beyond the operator
      lex_.next();
      e->setOperand(AstNode::RIGHT, parseExpression_());
      if (e->getOperand(AstNode::RIGHT) == 0) {
	error_("Missing expression");
      }
    } else {
      //... otherwise, there is either a ')' or an EOF condition (we are done!)
      e = term;
    }
    return e;
  }
    
  AstNode* parseParenExpression_() {
    AstNode* e = 0;
    // Advances the token cursor to the next token after '('
    lex_.next();
    e = parseExpression_();
    if (lex_.cur().getPartOfSpeech() != Token::RPAR) {
      error_("Missing )");
    }
    // The token cursor stays on top the last ')'
    return e;
  }

  AstNode::OpType parseOpType_() {
    // Expects to find the token cursor on top of the operator
    char ch = lex_.cur().getLexeme()[0];
    switch (ch) {
      case '+' : return AstNode::PLUS;
      case '-' : return AstNode::MINUS;
      case '*' : return AstNode::TIMES;
      case '/' : return AstNode::FRAC;
    default :
      return AstNode::NOP;
    }
  }

}; // End of class Parser

#endif
